//=============================================================================
//  MuseScore
//  Music Composition & Notation
//
//  Music Search
//
//=============================================================================
import QtQuick 2.2
import MuseScore 3.0
import FileIO 3.0

MuseScore {
	version: "1.0"
	description: "This plugin finds all equivalents of your selection in a score"
	menuPath: "Plugins.MusicSearch"

 	id: musicSearch

    Component.onCompleted : {
        if (mscoreMajorVersion >= 4) {
            musicSearch.title = "Music Search"
        }
     }

	FileIO {
            id: outfile
            source: tempPath() + "/ObjectExplorer.log"
            onError: console.log(msg)
      }  
	  // tempath() =  C:/Users/Dirk/AppData/Local/Temp/ObjectExplorer.log
      QProcess {
        id: proc
        }
            
      property variant output: "" 

	function doTheSearch(measure){
	
	function tick2measure(tick, measure){
		tick /= measure;
		tick++
		//avoid rounding
		tick -= 0.5;
		return tick.toFixed();	
	}
	
	function oneElement(cursor){
		var duration = cursor.element.duration.numerator / cursor.element.duration.denominator;
		duration *= 1920;
		if (cursor.element.tuplet) {
			var e = cursor.element;
			duration /= e.tuplet.actualNotes/e.tuplet.normalNotes;
		}
		return duration
	}
	
	var tpc2note = [
		[-1, "fbb"],
		[0, "cbb"],
		[1, "gbb"],
		[2, "dbb"],
		[3, "abb"],
		[4, "ebb"],
		[5, "bbb"],
		[6, "fb"],
		[7, "cb"],
		[8, "gb"],
		[9, "db"],
		[10, "ab"],
		[11, "eb"],
		[12, "bb"],
		[13, "f"],
		[14, "c"],
		[15, "g"],
		[16, "d"],
		[17, "a"],
		[18, "e"],
		[19, "b"],
		[20, "f#"],
		[21, "c#"],
		[22, "g#"],
		[23, "d#"],
		[24, "a#"],
		[25, "e#"],
		[26, "b#"],
		[27, "f##"],
		[28, "c##"],
		[29, "g##"],
		[30, "d##"],
		[31, "a##"],
		[32, "e##"],
		[33, "b##"],
	];
	
	function noteObject(tick, name, pitch, duration, type) {
		this.tick = tick;
		this.name = name;
		this.pitch = pitch;
		this.duration = duration;
		this.type = type;
	}
	
	function processFile(cursor){
		//var hticks = [];
		var tick = [];
		var name;
		var pitch;
		var data = [];
		var thisElement = []; // the container for objects
		
		// notenames code
      var fullScore = false;
		
		cursor.rewind(1);
		if (!cursor.segment) { // no selection
			fullScore = true;
			startStaff = 0; // start with 1st staff
			endStaff  = curScore.nstaves - 1; // and end with last
		} else {
			startStaff = cursor.staffIdx;
			cursor.rewind(2);
			 if (cursor.tick === 0) {
				// this happens when the selection includes
				// the last measure of the score.
				// rewind(2) goes behind the last segment (where
				// there's none) and sets tick=0
				endTick = curScore.lastSegment.tick + 1;
			 } else {
				endTick = cursor.tick;
			 }
			 endStaff = cursor.staffIdx;
		}
		console.log(startStaff + " - " + endStaff + " - " + endTick)
		
		// switch to fullscore qua ticks
		if (fullScore){
			endTick = curScore.lastSegment.tick + 1;
		}
		else{
			cursor.rewind(2);
			if (cursor.tick === 0) {
				// this happens when the selection includes
				// the last measure of the score.
				// rewind(2) goes behind the last segment (where
				// there's none) and sets tick=0
				endTick = curScore.lastSegment.tick + 1;
			} else {
				endTick = cursor.tick;
			}
			endStaff = cursor.staffIdx;
			cursor.rewind(1) // beginning of selection
		}
			
		console.log("cursor.tick = ", cursor.tick)	
			
			
			
		

		console.log(startStaff + " - " + endStaff + " - " + endTick)
		//console.time ("collect")
		var type
		// work backwards so that chords show lo to hi
			for (var track = ((endStaff + 1) * 4) -1; track >= startStaff; track--){		
				cursor.track = track;
				cursor.rewind(1); // beginning of selection
				cursor.track = track;
				if (fullScore)
					cursor.rewind(0); // beginning of score

					
				while (cursor.segment && (fullScore || cursor.tick < endTick)) {
					// collect notes, pitch, tick & duration
					if (cursor.element && cursor.element.type === Element.CHORD) {
						tick = cursor.tick					
						var notes = cursor.element.notes;
						if (notes.length > 1) type = "C" + notes.length
						else type = "N"
						
						for (var i = 0; i < notes.length; i++){
							var duration = oneElement(cursor)
							name = tpc2note[notes[i].tpc + 1][1]
							pitch = notes[i].pitch;
							// create a note object with above data
							// also for each note in a chord in the same track
							thisElement = new noteObject(tick, name, pitch, duration, type);
							data.push(thisElement); // push note object into data
						}
					} // end if CHORD
					else if (cursor.element && cursor.element.type === Element.REST) {
						// ignore rests	
					}// end if REST
					cursor.next();
				} // end while segment
			} // end for track	
			
			for (var j in data){
				//console.log ("720 j tick note,  = ",j + " " + data[j].tick + " " + data[j].name)
			}
			//output += "File: " + curScore.scoreName + "\n";
			if (nosel == true)
				output += "no selection made\n";
			// sort the data on tick & on pitch	
			data.sort(function(a, b) {
			  if (a.tick < b.tick) return -1;
			  if (a.tick > b.tick) return 1;
			  if (a.pitch < b.pitch) return -1;
			  if (a.pitch > b.pitch) return 1;
			  if (a.type < b.type) return -1;
			  if (a.type > b.type) return 1;
			  // Both identical
			  return 0;
			});
			
			for (var j in data){
				//console.log ("726 j tick note,  = ",j + " " + data[j].tick + " " + data[j].name + " " + chords[j].duration + " " + chords[j].type)
			}
		/*
		for (var i in data){
			console.log("data[i].name = ", i + " " + data[i].name + " " + data[i].ticks)
			console.log  ("duration = ", data[i].duration)
		}
		*/
		return data
	}
	
		var startStaff;
		var endStaff;
		var endTick;
		var fullScore = false;
		var cursor = curScore.newCursor();
		cursor.rewind(1); // beginning selection
		var selStartTick = cursor.tick;
		var nosel = false
	
		if (!cursor.segment) { // no selection
			console.log ("no selection made")
			//output += "File: " + curScore.scoreName + "\n";
			output += "no selection made\n";
			console.log ("output = ", output)
			nosel = true
			smartQuit();
		} else {
			startStaff = cursor.staffIdx;
			cursor.rewind(2);
			if (cursor.tick === 0) {
				// this happens when the selection includes
				// the last measure of the score.
				// rewind(2) goes behind the last segment (where
				// there's none) and sets tick=0
				endTick = curScore.lastSegment.tick + 1;
			} else {
				endTick = cursor.tick;
			}
			endStaff = cursor.staffIdx;
		}
		//var selEndTick = endTick;
		if (nosel == true)
			output += "no selection made\n";
		var needle = [];
		var flag = "NEEDLE" // or "FILE"
		
		needle = processFile(cursor,flag)
		
		
		var needlecount = 0
		for (var i in needle){
			console.log ("needle = ", needle[i].name + " " + needle[i].tick)
			needlecount++
		}
		console.log("needlecount = ", needlecount)
		curScore.startCmd();
        cmd("next-score");
        curScore.endCmd();
		var xdone = false
		while (xdone == false) {
			var ncount = needlecount
			cursor = curScore.newCursor()
			//output += "File: " + curScore.scoreName + "\n";
			if (nosel == true)
				output += "no selection made\n";
			var chords = [];
			flag = "FILE"
			chords = processFile(cursor,flag)
			
			for (var i in chords){
				//console.log ("chords = ", chords[i].name + " " + chords[i].ticks)
			}
			
			for (var i in chords){
				//console.log("chords[i].name = ", i + " " + chords[i].name + " " + chords[i].ticks)
				//console.log  ("duration = ", chords[i].duration)
			}
			
			
			for (var i in needle){
				//console.log ("249 needle = ", needle[i].name + " " + needle[i].tick)
			}
			console.log ("ncount = ", ncount)
			
			var done
			var n
			var match = false
			// go thru chords & loopk for a sequence that matches needle
			for (var i = 0; i < chords.length; i++) {
				n = 0;
				done = true
				// the needle sequence may be transposed
				// at the first note encountered store the pitch delta & the tick delta
				// subsequent notes must maintain the same deltas
				// pitchdelta
				var pdelta = 200
				// tickdelta
				var tdelta = -1
				while (done == true && n < ncount){
				if (i+n >= chords.length){
						done = false
						console.log ("269 i n ", i + " " + n)
					}
					if (i+n >= chords.length) done = false
					
					else {
						/*
						// debug version
						if (chords[i + n].duration != needle[n].duration){
							done = false
							console.log ("229 i n = ", i + " " + n)
						}
						if (chords[i + n].type  != needle[n].type){
							done = false
							console.log ("233 i n = ", i + " " + n)
						} 
						if (tdelta == -1) tdelta = needle[n].tick - chords[i + n].tick
						if (pdelta == 200) pdelta = (needle[n].pitch - chords[i + n].pitch)%12
						if (needle[n].tick - chords[i + n].tick != tdelta){
							done = false
							console.log ("239 i n = ", i + " " + n)
						}
						if (needle[n].pitch - chords[i + n].pitch != pdelta){
							done = false
							console.log ("243 i n = ", i + " " + n)
						} 
						*/
						///*
						// duration and type (note (N) or chord (C) must also match
						if (chords[i + n].duration != needle[n].duration) done = false
						if (chords[i + n].type  != needle[n].type) done = false
						if (tdelta == -1) tdelta = needle[n].tick - chords[i + n].tick
						if (pdelta == 200) pdelta = (needle[n].pitch - chords[i + n].pitch)%12
						if (needle[n].tick - chords[i + n].tick != tdelta) done = false
						if (needle[n].pitch - chords[i + n].pitch != pdelta) done = false
						//*/
					}
					n++;					
				}	
				if (done == true && n == ncount){
					// don't show the needle as match
					//if (chords[i].tick != needle[0].tick){
						var f0 = tick2measure(chords[i].tick, measure)
						if (match == false)
							output += "File: " + curScore.scoreName + "\n";
						var s = "equivalent found at measure " + f0
						output += s + "\n";
						match = true
						//console.log ("315 i n ", i + " " + n)
					//}
				}	
			}//i loop
			curScore.startCmd();
			cmd("next-score");
			curScore.endCmd();
			if(curScore.scoreName == "Selection") xdone = true
		}
		//console.time("write");
		var rc = outfile.write(output);
		if (rc){ 
			  console.log("Log file has been  written to "+outfile.source);
			  proc.start("notepad "+outfile.source); // Windows
		} else {
			  console.log("Something went wrong. File cannot be written");
		}
		//console.timeEnd("write");
		//console.timeEnd("start");
	} // end doTheSearch

	function smartQuit() {
		if (mscoreMajorVersion < 4) {Qt.quit()}
			else {quit()}
	}//smartQuit

	onRun: {
		var measure = 1920; // 1440 for 3/4; 1920 for 4/4
		curScore.startCmd();
		doTheSearch (measure);
		curScore.endCmd();
		smartQuit();  
		
	}
}